// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file MeshDistributor.h */

#ifndef AMDIS_MESHDISTRIBUTOR_H
#define AMDIS_MESHDISTRIBUTOR_H


#include <map>
#include <set>
#include <vector>
#include <mpi.h>

#include "Global.h"
#include "ProblemTimeInterface.h"
#include "ProblemIterationInterface.h"
#include "FiniteElemSpace.h"
#include "AdaptInfo.h"
#include "parallel/InteriorBoundary.h"
#include "Serializer.h"
#include "BoundaryManager.h"
#include "ElementObjectData.h"
#include "AMDiS_fwd.h"

namespace AMDiS {
  
  class ParMetisPartitioner;

  class MeshDistributor
  {
  protected:
    /// Defines a mapping type from DOFs to rank numbers.
    typedef std::map<const DegreeOfFreedom*, int> DofToRank;

    /// Defines a mapping type from DOFs to a set of rank numbers.
    typedef std::map<const DegreeOfFreedom*, std::set<int> > DofToPartitions;

    /// Defines a mapping type from rank numbers to sets of DOFs.
    typedef std::map<int, DofContainer> RankToDofContainer;

    /// Defines a mapping type from DOF indices to DOF indices.
    typedef std::map<DegreeOfFreedom, DegreeOfFreedom> DofMapping;

    /// Defines a mapping type from DOFs to boolean values.
    typedef std::map<const DegreeOfFreedom*, bool> DofToBool;

    /// Defines a mapping type from DOF indices to boolean values.
    typedef std::map<DegreeOfFreedom, bool> DofIndexToBool;

    /// Forward type (it maps rank numbers to the interior boundary objects).
    typedef InteriorBoundary::RankToBoundMap RankToBoundMap;

    typedef std::map<const DegreeOfFreedom*, DegreeOfFreedom> DofIndexMap;

    /// Mapps a boundar type, i.e., a boundary identifier index, to a periodic 
    /// dof mapping.
    typedef std::map<BoundaryType, DofMapping> PeriodicDofMap;
    
    typedef std::vector<MeshStructure> MeshCodeVec;

  public:
    MeshDistributor(std::string str);
		          
    virtual ~MeshDistributor() {}

    void initParallelization(AdaptInfo *adaptInfo);

    void exitParallelization(AdaptInfo *adaptInfo);

    void addProblemStat(ProblemVec *probVec);

    /** \brief
     * This function checks if the mesh has changed on at least on rank. In this case,
     * the interior boundaries are adapted on all ranks such that they fit together on
     * all ranks. Furthermore the function \ref updateLocalGlobalNumbering() is called
     * to update the dof numberings and mappings on all rank due to the new mesh
     * structure.
     */
    void checkMeshChange();

    void updateDofAdmins();    

    /** \brief
     * Test, if the mesh consists of macro elements only. The mesh partitioning of
     * the parallelization works for macro meshes only and would fail, if the mesh
     * is already refined in some way. Therefore, this function will exit the program
     * if it finds a non macro element in the mesh.
     */
    void testForMacroMesh();

    /// Set for each element on the partitioning level the number of leaf elements.
    double setElemWeights(AdaptInfo *adaptInfo);

    void partitionMesh(AdaptInfo *adaptInfo);

    inline virtual std::string getName() 
    { 
      return name; 
    }

    /// Returns \ref feSpace.
    inline const FiniteElemSpace* getFeSpace()
    {
      return feSpace;
    }
    
    /// Returns \ref nRankDOFs, the number of DOFs in the rank mesh.
    inline int getNumberRankDofs() 
    {
      return nRankDofs;
    }

    /// Returns \ref nOverallDofs, the global number of DOFs.
    inline int getNumberOverallDofs()
    {
      return nOverallDofs;
    }

    /// Maps a local dof to its global index.
    inline DegreeOfFreedom mapLocalToGlobal(DegreeOfFreedom dof)
    {
      return mapLocalGlobalDofs[dof];
    }

    /// Maps a local dof to its local index.
    inline DegreeOfFreedom mapLocalToDofIndex(DegreeOfFreedom dof)
    {
      return mapLocalDofIndex[dof];
    }

    /// Returns for a global dof index its periodic mapping for a given boundary type.
    inline int getPeriodicMapping(BoundaryType type, int globalDofIndex)
    {
      TEST_EXIT_DBG(periodicDof[type].count(globalDofIndex) == 1)
	("Should not happen!\n");

      return periodicDof[type][globalDofIndex];
    }

    /// For a given global DOF index, this function returns the set of periodic
    /// associations, i.e., the boundary types the DOF is associated to, for this DOF.
    inline std::set<BoundaryType>& getPerDofAssociations(int globalDofIndex)
    {      
      return periodicDofAssociations[globalDofIndex];
    }

    /// Returns true, if the DOF (global index) is a periodic DOF.
    inline bool isPeriodicDof(int globalDofIndex)
    {
      return (periodicDofAssociations.count(globalDofIndex) > 0);
    }

    /// Returns true, if the DOF (global index) is a periodic DOF for the given
    /// boundary type.
    inline bool isPeriodicDof(int globalDofIndex, BoundaryType type)
    {
      return (periodicDof[type].count(globalDofIndex) > 0);
    }

    /// Return true, if the given DOF is owned by the rank. If false, the DOF is in
    /// rank's partition, but is owned by some other rank.
    inline bool getIsRankDof(DegreeOfFreedom dof)
    {
      return isRankDof[dof];
    }

    inline long getLastMeshChangeIndex()
    {
      return lastMeshChangeIndex;
    }

    inline int getRstart()
    {
      return rstart;
    }

    inline int getMpiRank()
    {
      return mpiRank;
    }

    inline MPI::Intracomm& getMpiComm()
    {
      return mpiComm;
    }

    inline RankToDofContainer& getSendDofs()
    {
      return sendDofs;
    }

    inline RankToDofContainer& getRecvDofs()
    {
      return recvDofs;
    }

    // Writes all data of this object to an output stream.
    void serialize(std::ostream &out);

    // Reads the object data from an input stream.
    void deserialize(std::istream &in);

    /** \brief
     * This function must be used if the values of a DOFVector must be synchronised
     * over all ranks. That means, that each rank sends the values of the DOFs, which
     * are owned by the rank and lie on an interior bounday, to all other ranks also
     * having these DOFs.
     *
     * This function must be used, for example, after the lineary system is solved, or
     * after the DOFVector is set by some user defined functions, e.g., initial
     * solution functions.
     */    
    void synchVector(DOFVector<double> &vec);

    /** \brief
     * Works in the same way as the function above defined for DOFVectors. Due to
     * performance, this function does not call \ref synchVector for each DOFVector,
     * but instead sends all values of all DOFVectors all at once.
     */
    void synchVector(SystemVector &vec);

  protected:
    /** \brief
     * Determines the interior boundaries, i.e. boundaries between ranks, and stores
     * all information about them in \ref interiorBoundary.
     */
    void createInteriorBoundaryInfo();

    void createBoundaryDataStructure();

    /// Removes all macro elements from the mesh that are not part of ranks partition.
    void removeMacroElements();

    /// Creates from a macro mesh a correct local and global DOF index numbering.
    void createLocalGlobalNumbering();

    /// Updates the local and global DOF numbering after the mesh has been changed.
    void updateLocalGlobalNumbering();

    /** \brief
     * Creates to all dofs in rank's partition that are on a periodic boundary the
     * mapping from dof index to the other periodic dof indices. This information
     * is stored in \ref periodicDof.
     */
    void createPeriodicMap();

    /** \brief
     * This function create new mappings from local to global indices, 
     * \ref mapLocalGlobalDofs, and from local to dof indices, \ref mapLocalDofIndex.
     * Furthermore, using the first argument the dof indices in ranks partition are
     * changed.
     * 
     * \param[in] rankDofsNewLocalIndex       Map from dof pointers of all dofs in rank
     *                                        to new dof indices.
     * \param[in] rankOwnedDofsNewLocalIndex  Map from dof pointers of dofs owned by
     *                                        the rank to the new local index.
     * \param[in] rankDofsNewGlobalIndex      Map from dof pointers of all dofs in rank
     *                                        to the new global index.
     */
    void createLocalMappings(DofIndexMap &rankDofsNewLocalIndex,
			     DofIndexMap &rankOwnedDofsNewLocalIndex,
			     DofIndexMap &rankDofsNewGlobalIndex);

    /** \brief
     * This function traverses the whole mesh, i.e. before it is really partitioned,
     * and collects information about which DOF corresponds to which rank. Can only
     * be used, if \ref partitionVec is set correctly. This is only the case, when
     * the macro mesh is partitioned.
     *
     * \param[out] partitionDOFs   Stores to each DOF pointer the set of ranks the DOF
     *                             is part of.
     * \param[out] rankDOFs        Stores all rank DOFs.
     * \param[out] boundaryDOFs    Stores all DOFs in ranks partition that are on an 
     *                             interior boundary but correspond to another rank.
     */
    void createDofMemberInfo(DofToPartitions& partitionDofs,
			     DofContainer& rankOwnedDofs,
			     DofContainer& rankAllDofs,
			     DofToRank& boundaryDofs);

    /** \brief
     * Checks for all given interior boundaries if the elements fit together on both
     * sides of the boundaries. If this is not the case, the mesh is adapted. Because
     * refinement of a certain element may forces the refinement of other elements,
     * it is not guaranteed that all rank's meshes fit together after this function
     * terminates. Hence, it must be called until a stable mesh refinement is reached.
     * If the mesh has  been changed by this function, it returns true. Otherwise, it
     * returns false, i.e., the given interior boundaries fit together on both sides.
     *
     * \param[in] allBound   Defines a map from rank to interior boundaries which 
     *                       should be checked.
     */
    bool checkAndAdaptBoundary(RankToBoundMap &allBound);
    
    /** \brief
     * This functions create a Paraview file with the macro mesh where the elements
     * are colored by the partition they are part of. This function can be used for
     * debugging.
     */
    void writePartitioningMesh(std::string filename);

    /// Sets \ref isRankDof to all matrices and rhs vectors in all stationary problems.
    void setRankDofs();

    /// Removes all periodic boundary condition information from all matrices and
    /// vectors of all stationary problems and from the mesh itself.
    void removePeriodicBoundaryConditions();

    // Removes all periodic boundaries from a given boundary map.
    void removePeriodicBoundaryConditions(BoundaryIndexMap& boundaryMap);

    /** \brief
     * Starts the procedure to fit a given edge/face of an element with a mesh
     * structure code. This functions prepares some data structures and call
     * then \ref fitElementToMeshCode, that mainly refines the element such that
     * it fits to the mesh structure code.
     *
     * \param[in] code         The mesh structure code to which the edge/face of
     *                         an element must be fitted.
     * \param[in] el           Pointer to the element.
     * \param[in] subObj       Defines whether an edge or a face must be fitted.
     * \param[in] ithObj       Defines which edge/face must be fitted.
     * \param[in] elType       Element type of the element (only important in 3D).
     * \param[in] reverseMode  Defines, whether the mesh structure code is given
     *                         in reverse mode, i.e., left and right children where
     *                         changed when the code was created.
     */
    bool startFitElementToMeshCode(MeshStructure &code, 
				   Element *el, 
				   GeoIndex subObj,
				   int ithObj, 
				   int elType,
				   bool reverseMode);
    
    /** \brief
     * Recursively fits a given mesh structure code to an edge/face of an element.
     * This function is always initialy called from \ref startFitElementToMeshCode.
     *
     * \param[in] code         The mesh structure code which is used to fit an
     *                         edge/face of an element.
     * \param[in] stack        A traverse stack object. The upper most element in this
     *                         stack must be used for fitting the mesh structure code
     *                         at the current position.
     * \param[in] subObj       Defines whether an edge or a face must be fitted.
     * \param[in] ithObj       Defines which edge/face must be fitted.
     * \param[in] reverseMode  Defines, whether the mesh structure code is given
     *                         in reverse mode, i.e., left and right children where
     *                         changed when the code was created.
     */
    bool fitElementToMeshCode(MeshStructure &code, 
			      TraverseStack &stack,
			      GeoIndex subObj,
			      int ithObj,
			      bool reverseMode);

    /// Writes a vector of dof pointers to an output stream.
    void serialize(std::ostream &out, DofContainer &data);

    /// Reads a vector of dof pointers from an input stream.
    void deserialize(std::istream &in, DofContainer &data,
		     std::map<int, const DegreeOfFreedom*> &dofMap);

    /// Writes a \ref RankToDofContainer to an output stream.
    void serialize(std::ostream &out, RankToDofContainer &data);

    /// Reads a \ref RankToDofContainer from an input stream.
    void deserialize(std::istream &in, RankToDofContainer &data,
		     std::map<int, const DegreeOfFreedom*> &dofMap);

    /// Writes a periodic dof mapping to an output stream.
    void serialize(std::ostream &out, PeriodicDofMap &data);

    void serialize(std::ostream &out, std::map<int, std::set<int> >& data);

    /// Reads a periodic dof mapping from an input stream.
    void deserialize(std::istream &in, PeriodicDofMap &data);

    void deserialize(std::istream &in, std::map<int, std::set<int> >& data);

    /// Writes a mapping from dof pointers to some values to an output stream.
    template<typename T>
    void serialize(std::ostream &out, std::map<const DegreeOfFreedom*, T> &data)
    {
      FUNCNAME("ParallelDomainBase::serialize()");

      int mapSize = data.size();
      SerUtil::serialize(out, mapSize);
      for (typename std::map<const DegreeOfFreedom*, T>::iterator it = data.begin();
	   it != data.end(); ++it) {
	int v1 = (*(it->first));
	T v2 = it->second;
	SerUtil::serialize(out, v1);
	SerUtil::serialize(out, v2);
      }
    }

    /// Reads a mapping from dof pointer to some values from an input stream.
    template<typename T>
    void deserialize(std::istream &in, std::map<const DegreeOfFreedom*, T> &data,
		     std::map<int, const DegreeOfFreedom*> &dofMap)
    {
      FUNCNAME("ParallelDomainBase::deserialize()");

      int mapSize = 0;
      SerUtil::deserialize(in, mapSize);
      for (int i = 0; i < mapSize; i++) {
	int v1 = 0;
	T v2;
	SerUtil::deserialize(in, v1);
	SerUtil::deserialize(in, v2);

	TEST_EXIT_DBG(dofMap.count(v1) != 0)("Cannot find DOF %d in map!\n", v1);

	data[dofMap[v1]] = v2;
      }
    }
		        
  protected:
    ///
    std::vector<ProblemVec*> probStat;

    /// The rank of the current process.
    int mpiRank;

    /// Overall number of processes.
    int mpiSize;

    /** \brief
     * MPI communicator collected all processes, which should
     * be used for calculation. The Debug procces is not included
     * in this communicator.
     */
    MPI::Intracomm mpiComm;

    /// Name of the problem (as used in the init files)
    std::string name;

    /// Finite element space of the problem.
    FiniteElemSpace *feSpace;

    /// Mesh of the problem.
    Mesh *mesh;

    /** \brief
     * A refinement manager that should be used on the mesh. It is used to refine
     * elements at interior boundaries in order to fit together with elements on the
     * other side of the interior boundary.
     */    
    RefinementManager *refineManager;

    /// Info level.
    int info;

    /// Pointer to the paritioner which is used to devide a mesh into partitions.
    ParMetisPartitioner *partitioner;

    /// Weights for the elements, i.e., the number of leaf elements within this element.
    std::map<int, double> elemWeights;

    /// Is true, if the mesh was not partitioned before, otherwise it's false.
    bool initialPartitionMesh;

    /** \brief
     * Stores to every coarse element index the number of the partition it 
     * corresponds to.
     */
    std::map<int, int> partitionVec;

    /** \brief
     * Stores an old partitioning of elements. To every element index the number
     * of the parition it corresponds to is stored.
     */
    std::map<int, int> oldPartitionVec;    
   
    /// Number of DOFs in the rank mesh.
    int nRankDofs;

    /// Number of DOFs in the whole domain.
    int nOverallDofs;

    /** \brief 
     * Defines the interior boundaries of the domain that result from partitioning
     * the whole mesh. Contains only the boundaries, which are owned by the rank, i.e.,
     * the object gives for every neighbour rank i the boundaries this rank owns and 
     * shares with rank i.
     */
    InteriorBoundary myIntBoundary;
    
    /** \brief
     * Defines the interior boundaries of the domain that result from partitioning
     * the whole mesh. Contains only the boundaries, which are not owned by the rank,
     * i.e., the object gives for every neighbour rank i the boundaries that are
     * owned by rank i and are shared with this rank.
     */
    InteriorBoundary otherIntBoundary;

    /** \brief
     * Defines the periodic boundaries with other ranks. Periodic boundaries have
     * no owner, as it is the case of interior boundaries.
     */
    InteriorBoundary periodicBoundary;

    /** \brief
     * This map contains for each rank the list of dofs the current rank must send
     * to exchange solution dofs at the interior boundaries.
     */
    RankToDofContainer sendDofs;

    /** \brief
     * This map contains for each rank the list of DOFs from which the current rank 
     * will receive DOF values (i.e., this are all DOFs at an interior boundary). The
     * DOF indices are given in rank's local numbering.
     */
    RankToDofContainer recvDofs;

    /// Maps local to global dof indices.
    DofMapping mapLocalGlobalDofs;

    /// Maps local dof indices to real dof indices.
    DofMapping mapLocalDofIndex;  

    /** \brief
     * Maps all DOFs in ranks partition to a bool value. If it is true, the DOF is
     * owned by the rank. Otherwise, its an interior boundary DOF that is owned by
     * another rank.
     */
    DofIndexToBool isRankDof;

    /** \brief
     * If periodic boundaries are used, this map stores, for each periodic boundary
     * type, for all DOFs in rank's partition (that are on periodic boundaries), the 
     * corresponding mapped periodic DOFs. The mapping is defined by using global 
     * dof indices.
     */
    PeriodicDofMap periodicDof;
    
    /** \brief
     * If periodic boundaries are used, this map stores to each periodic DOF in rank's
     * partition the set of periodic boundaries the DOF is associated to. In 2D, most
     * DOFs are only on one periodic boundary. Only, e.g., in a box with all boundaries
     * being periodic, the for corners are associated by two different boundaries.     
     */
    std::map<int, std::set<BoundaryType> > periodicDofAssociations;

    /// Is the index of the first row of the linear system, which is owned by the rank.
    int rstart;

    /** \brief
     * If the problem definition has been read from a serialization file, this 
     * variable is true, otherwise it is false. This variable is used to stop the
     * initialization function, if the problem definition has already been read from
     * a serialization file.
     */
    bool deserialized;

    /// Denotes whether there exists a filewriter for this object.
    bool writeSerializationFile;

    /** \brief
     * Stores the mesh change index. This is used to recognize changes in the mesh 
     * structure (e.g. through refinement or coarsening managers).
     */
    long lastMeshChangeIndex;

    /// This variable is true, if the macro elements are consistent with all other
    /// data structures. Within the initialization and during mesh redistribution this
    /// may not be the case.
    bool macroElementStructureConsisten;

    friend class ParallelDebug;
  };
}

#endif // AMDIS_MESHDISTRIBUTOR_H
